/*
  This code was written by Xiaokui Xiao.
  Please read the following.

    Raymond Chi-Wing Wong, Yufei Tao, Ada Wai-Chee Fu and Xiaokui Xiao. 
    "On Efficient Spatial Matching", 
    the 33rd International Conference on Very Large Data Bases (VLDB'07), 
    Vienna, Austria on 23-28 Sept, 2007 

*/

#include "ClosestPair.h"

#include <cmath>
#include <queue>
#include <vector>

XkPair XkFirstPair( node_type* pRootO, node_type* pRootP, rtree_info* aInfoO, rtree_info* aInfoP )
{
	using namespace std;

	int	i, j, k, ii, jj;
	node_type* p_node_O = pRootO;
	node_type* p_node_P = pRootP;

	node_type* p_child_O;
	node_type* p_child_P;

	int	num_O, num_P;
	float dist, dist1, dist2, dist3, overlap;
	float min_dist;
	float max_overlap;

	XkPair rslt_pair;

	while (true)
	{
		if ((p_node_O->attribute == LEAF) && (p_node_P->attribute == LEAF))
		{
			return rslt_pair;
		}

		if (p_node_O->attribute == LEAF)
		{
			num_O = 1;
		}
		else
		{
			num_O = aInfoO->M - p_node_O->vacancy;
		}

		if (p_node_P->attribute == LEAF)
		{
			num_P = 1;
		}
		else
		{
			num_P = aInfoP->M - p_node_P->vacancy;
		}

		min_dist = (float) 10e10;
		max_overlap = -1;

		for (i = 0; i < num_O; ++i)
		{
			if (p_node_O->attribute == LEAF)
			{
				p_child_O = p_node_O;
			}
			else
			{
				p_child_O = p_node_O->ptr[i];
			}			

			for (j = 0; j < num_P; ++j)
			{
				if (p_node_P->attribute == LEAF)
				{
					p_child_P = p_node_P;
				}
				else
				{
					p_child_P = p_node_P->ptr[j];
				}

				dist = 0;
				overlap = 1;
				for (k = 0; k < aInfoO->dim; ++k)
				{
					dist1 = abs( p_child_O->a[k] - p_child_P->b[k] );
					dist2 = abs( p_child_O->b[k] - p_child_P->a[k] );
					dist3 = (dist1 < dist2 ? dist1 : dist2);
					if ((p_child_O->a[k] > p_child_P->b[k]) || (p_child_O->b[k] < p_child_P->a[k]))
					{						
						dist += dist3 * dist3;
						overlap = 0;
					}
					else
					{
						overlap *= dist3;
					}
				}
				if (dist < min_dist - 10e-6)
				{
					min_dist = dist;
					max_overlap = overlap;
					ii = i;
					jj = j;
				}
				else if (dist < min_dist + 10e-6)
				{
					if (overlap > max_overlap)
					{
						max_overlap = overlap;
						ii = i;
						jj = j;
					}
				}
			}
		}

		if (p_node_O->attribute == LEAF)
		{
			rslt_pair.m_pNodeO = p_node_O;			
		}
		else
		{
			rslt_pair.m_pNodeO = p_node_O = p_node_O->ptr[ii];
		}

		if (p_node_P->attribute == LEAF)
		{
			rslt_pair.m_pNodeP = p_node_P;
		}
		else
		{
			rslt_pair.m_pNodeP = p_node_P = p_node_P->ptr[jj];
		}
		
		rslt_pair.m_fDist = min_dist;
	}
}


XkPair XkClosestPair( node_type* pRootO, node_type* pRootP, rtree_info* aInfoO, rtree_info* aInfoP )
{
	using namespace std;

	int	i, j, k;
	node_type* p_node_O = pRootO;
	node_type* p_node_P = pRootP;

	node_type* p_child_O;
	node_type* p_child_P;

	int	num_O, num_P;
	float dist, dist1, dist2, dist3;
	float min_dist = (float) 10e10;
	
	XkPair temp_pair;
	XkPair rslt_pair;

	priority_queue< XkPair, vector<XkPair>, greater<XkPair> > heap;

	rslt_pair = XkFirstPair( pRootO, pRootP, aInfoO, aInfoP );
	min_dist = rslt_pair.m_fDist;

	temp_pair.m_pNodeO = pRootO;
	temp_pair.m_pNodeP = pRootP;
	temp_pair.m_fDist = -1;
	heap.push( temp_pair );

	while (heap.empty() == false)
	{
		temp_pair = heap.top();
		heap.pop();

		if (min_dist < temp_pair.m_fDist)
		{
			return rslt_pair;
		}

		if ((temp_pair.m_pNodeO->attribute == LEAF) && (temp_pair.m_pNodeP->attribute == LEAF))
		{
			if (temp_pair.m_fDist < min_dist)
			{
				min_dist  = temp_pair.m_fDist;
				rslt_pair = temp_pair;
				continue;
			}
		}
		else
		{
			p_node_O = temp_pair.m_pNodeO;
			p_node_P = temp_pair.m_pNodeP;

			if (p_node_O->attribute == LEAF)
			{
				num_O = 1;
			}
			else
			{
				num_O = aInfoO->M - p_node_O->vacancy;
			}

			if (p_node_P->attribute == LEAF)
			{
				num_P = 1;
			}
			else
			{
				num_P = aInfoP->M - p_node_P->vacancy;
			}

			for (i = 0; i < num_O; ++i)
			{
				if (p_node_O->attribute == LEAF)
				{
					p_child_O = p_node_O;
				}
				else
				{
					p_child_O = p_node_O->ptr[i];
				}			

				for (j = 0; j < num_P; ++j)
				{				
					if (p_node_P->attribute == LEAF)
					{
						p_child_P = p_node_P;
					}
					else
					{
						p_child_P = p_node_P->ptr[j];
					}					
					
					dist = 0;
					for (k = 0; k < aInfoO->dim; ++k)
					{
						if ((p_child_O->a[k] > p_child_P->b[k]) || (p_child_O->b[k] < p_child_P->a[k]))
						{						
							dist1 = abs( p_child_O->a[k] - p_child_P->b[k] );
							dist2 = abs( p_child_O->b[k] - p_child_P->a[k] );
							dist3 = (dist1 < dist2 ? dist1 : dist2);
							dist += dist3 * dist3;
						}
					}

					if (dist < min_dist - 10e-6)
					{
						temp_pair.m_pNodeO = p_child_O;
						temp_pair.m_pNodeP = p_child_P;
						temp_pair.m_fDist = dist;
						if ((temp_pair.m_pNodeO->attribute == LEAF) && (temp_pair.m_pNodeP->attribute == LEAF))
						{
							min_dist = dist;
							rslt_pair = temp_pair;
						}
						heap.push( temp_pair );
					}
				}
			}			
		}
	}

	return rslt_pair;
}


